"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.LocalDetachedFluidDataStoreContext = exports.LocalFluidDataStoreContext = exports.LocalFluidDataStoreContextBase = exports.RemoteFluidDataStoreContext = exports.FluidDataStoreContext = exports.createAttributesBlob = void 0;
const container_definitions_1 = require("@fluidframework/container-definitions");
const client_utils_1 = require("@fluid-internal/client-utils");
const core_utils_1 = require("@fluidframework/core-utils");
const driver_utils_1 = require("@fluidframework/driver-utils");
const runtime_definitions_1 = require("@fluidframework/runtime-definitions");
const runtime_utils_1 = require("@fluidframework/runtime-utils");
const telemetry_utils_1 = require("@fluidframework/telemetry-utils");
const index_js_1 = require("./summary/index.js");
const index_js_2 = require("./gc/index.js");
function createAttributes(pkg, isRootDataStore) {
    const stringifiedPkg = JSON.stringify(pkg);
    return {
        pkg: stringifiedPkg,
        summaryFormatVersion: 2,
        isRootDataStore,
    };
}
function createAttributesBlob(pkg, isRootDataStore) {
    const attributes = createAttributes(pkg, isRootDataStore);
    return new driver_utils_1.BlobTreeEntry(index_js_1.dataStoreAttributesBlobName, JSON.stringify(attributes));
}
exports.createAttributesBlob = createAttributesBlob;
/**
 * Represents the context for the store. This context is passed to the store runtime.
 */
class FluidDataStoreContext extends client_utils_1.TypedEventEmitter {
    get packagePath() {
        (0, core_utils_1.assert)(this.pkg !== undefined, 0x139 /* "Undefined package path" */);
        return this.pkg;
    }
    get options() {
        return this.parentContext.options;
    }
    get clientId() {
        return this.parentContext.clientId;
    }
    get clientDetails() {
        return this.parentContext.clientDetails;
    }
    get logger() {
        return this.parentContext.logger;
    }
    get deltaManager() {
        return this.parentContext.deltaManager;
    }
    get connected() {
        return this.parentContext.connected;
    }
    get IFluidHandleContext() {
        return this.parentContext.IFluidHandleContext;
    }
    get containerRuntime() {
        return this._containerRuntime;
    }
    ensureNoDataModelChanges(callback) {
        return this.parentContext.ensureNoDataModelChanges(callback);
    }
    get isLoaded() {
        return this.loaded;
    }
    get baseSnapshot() {
        return this._baseSnapshot;
    }
    get idCompressor() {
        return this.parentContext.idCompressor;
    }
    get disposed() {
        return this._disposed;
    }
    get tombstoned() {
        return this._tombstoned;
    }
    get attachState() {
        return this._attachState;
    }
    get IFluidDataStoreRegistry() {
        (0, core_utils_1.assert)(this.channel !== undefined, 0x8f3 /* This should be called after the channel is created, when the registry is populated */);
        return this.registry;
    }
    /**
     * A datastore is considered as root if it
     * 1. is root in memory - see isInMemoryRoot
     * 2. is root as part of the base snapshot that the datastore loaded from
     * @returns whether a datastore is root
     */
    async isRoot(aliasedDataStores) {
        if (this.isInMemoryRoot()) {
            return true;
        }
        // This if is a performance optimization.
        // We know that if the base snapshot is omitted, then the isRootDataStore flag is not set.
        // That means we can skip the expensive call to getInitialSnapshotDetails for virtualized datastores,
        // and get the information from the alias map directly.
        if (aliasedDataStores !== undefined && this.baseSnapshot?.omitted === true) {
            return aliasedDataStores.has(this.id);
        }
        return (await this.getInitialSnapshotDetails()).isRootDataStore;
    }
    /**
     * There are 3 states where isInMemoryRoot needs to be true
     * 1. when a datastore becomes aliased. This can happen for both remote and local datastores
     * 2. when a datastore is created locally as root
     * 3. when a datastore is created locally as root and is rehydrated
     * @returns whether a datastore is root in memory
     */
    isInMemoryRoot() {
        return this._isInMemoryRoot;
    }
    constructor(props, existing, isLocalDataStore, makeLocallyVisibleFn) {
        super();
        this.existing = existing;
        this.isLocalDataStore = isLocalDataStore;
        this.makeLocallyVisibleFn = makeLocallyVisibleFn;
        this._disposed = false;
        /**
         * A Tombstoned object has been unreferenced long enough that GC knows it won't be referenced again.
         * Tombstoned objects are eventually deleted by GC.
         */
        this._tombstoned = false;
        /** If true, this means that this data store context and its children have been removed from the runtime */
        this.deleted = false;
        this.detachedRuntimeCreation = false;
        this.loaded = false;
        this.pending = [];
        this._isInMemoryRoot = false;
        this._containerRuntime = props.parentContext.containerRuntime;
        this.parentContext = props.parentContext;
        this.id = props.id;
        this.storage = props.storage;
        this.scope = props.scope;
        this.pkg = props.pkg;
        this.loadingGroupId = props.loadingGroupId;
        // URIs use slashes as delimiters. Handles use URIs.
        // Thus having slashes in types almost guarantees trouble down the road!
        (0, core_utils_1.assert)(!this.id.includes("/"), 0x13a /* Data store ID contains slash */);
        this._attachState =
            this.parentContext.attachState !== container_definitions_1.AttachState.Detached && this.existing
                ? this.parentContext.attachState
                : container_definitions_1.AttachState.Detached;
        const thisSummarizeInternal = async (fullTree, trackState, telemetryContext) => this.summarizeInternal(fullTree, trackState, telemetryContext);
        this.summarizerNode = props.createSummarizerNodeFn(thisSummarizeInternal, async (fullGC) => this.getGCDataInternal(fullGC));
        this.mc = (0, telemetry_utils_1.createChildMonitoringContext)({
            logger: this.logger,
            namespace: "FluidDataStoreContext",
            properties: {
                all: (0, telemetry_utils_1.tagCodeArtifacts)({
                    fluidDataStoreId: this.id,
                    // The package name is a getter because `this.pkg` may not be initialized during construction.
                    // For data stores loaded from summary, it is initialized during data store realization.
                    fullPackageName: () => this.pkg?.join("/"),
                }),
            },
        });
        this.thresholdOpsCounter = new telemetry_utils_1.ThresholdCounter(FluidDataStoreContext.pendingOpsCountThreshold, this.mc.logger);
        this.gcThrowOnTombstoneUsage = this.parentContext.gcThrowOnTombstoneUsage;
        this.gcTombstoneEnforcementAllowed = this.parentContext.gcTombstoneEnforcementAllowed;
        // By default, a data store can log maximum 10 local changes telemetry in summarizer.
        this.localChangesTelemetryCount =
            this.mc.config.getNumber("Fluid.Telemetry.LocalChangesTelemetryCount") ?? 10;
    }
    dispose() {
        if (this._disposed) {
            return;
        }
        this._disposed = true;
        // Dispose any pending runtime after it gets fulfilled
        // Errors are logged where this.channelP is consumed/generated (realizeCore(), bindRuntime())
        if (this.channelP) {
            this.channelP
                .then((runtime) => {
                runtime.dispose();
            })
                .catch((error) => { });
        }
    }
    /**
     * When delete is called, that means that the data store is permanently removed from the runtime, and will not show up in future summaries
     * This function is called to prevent ops from being generated from this data store once it has been deleted. Furthermore, this data store
     * should not receive any ops/signals.
     */
    delete() {
        this.deleted = true;
    }
    setTombstone(tombstone) {
        if (this.tombstoned === tombstone) {
            return;
        }
        this._tombstoned = tombstone;
    }
    rejectDeferredRealize(reason, failedPkgPath, fullPackageName) {
        throw new telemetry_utils_1.LoggingError(reason, (0, telemetry_utils_1.tagCodeArtifacts)({
            failedPkgPath,
            packagePath: fullPackageName?.join("/"),
        }));
    }
    async realize() {
        (0, core_utils_1.assert)(!this.detachedRuntimeCreation, 0x13d /* "Detached runtime creation on realize()" */);
        if (!this.channelP) {
            this.channelP = this.realizeCore(this.existing).catch((error) => {
                const errorWrapped = telemetry_utils_1.DataProcessingError.wrapIfUnrecognized(error, "realizeFluidDataStoreContext");
                errorWrapped.addTelemetryProperties((0, telemetry_utils_1.tagCodeArtifacts)({
                    fullPackageName: this.pkg?.join("/"),
                    fluidDataStoreId: this.id,
                }));
                this.mc.logger.sendErrorEvent({ eventName: "RealizeError" }, errorWrapped);
                throw errorWrapped;
            });
        }
        return this.channelP;
    }
    async factoryFromPackagePath() {
        const packages = this.pkg;
        if (packages === undefined) {
            this.rejectDeferredRealize("packages is undefined");
        }
        let entry;
        let registry = this.parentContext.IFluidDataStoreRegistry;
        let lastPkg;
        for (const pkg of packages) {
            if (!registry) {
                this.rejectDeferredRealize("No registry for package", lastPkg, packages);
            }
            lastPkg = pkg;
            entry = await registry.get(pkg);
            if (!entry) {
                this.rejectDeferredRealize("Registry does not contain entry for the package", pkg, packages);
            }
            registry = entry.IFluidDataStoreRegistry;
        }
        const factory = entry?.IFluidDataStoreFactory;
        if (factory === undefined) {
            this.rejectDeferredRealize("Can't find factory for package", lastPkg, packages);
        }
        (0, core_utils_1.assert)(this.registry === undefined, 0x157 /* "datastore registry already attached" */);
        this.registry = registry;
        return factory;
    }
    async realizeCore(existing) {
        const details = await this.getInitialSnapshotDetails();
        // Base snapshot is the baseline where pending ops are applied to.
        // It is important that this be in sync with the pending ops, and also
        // that it is set here, before bindRuntime is called.
        this._baseSnapshot = details.snapshot;
        this.baseSnapshotSequenceNumber = details.sequenceNumber;
        (0, core_utils_1.assert)(this.pkg === details.pkg, 0x13e /* "Unexpected package path" */);
        const factory = await this.factoryFromPackagePath();
        const channel = await factory.instantiateDataStore(this, existing);
        (0, core_utils_1.assert)(channel !== undefined, 0x140 /* "undefined channel on datastore context" */);
        await this.bindRuntime(channel, existing);
        // This data store may have been disposed before the channel is created during realization. If so,
        // dispose the channel now.
        if (this.disposed) {
            channel.dispose();
        }
        return channel;
    }
    /**
     * Notifies this object about changes in the connection state.
     * @param value - New connection state.
     * @param clientId - ID of the client. Its old ID when in disconnected state and
     * its new client ID when we are connecting or connected.
     */
    setConnectionState(connected, clientId) {
        // ConnectionState should not fail in tombstone mode as this is internally run
        this.verifyNotClosed("setConnectionState", false /* checkTombstone */);
        // Connection events are ignored if the store is not yet loaded
        if (!this.loaded) {
            return;
        }
        (0, core_utils_1.assert)(this.connected === connected, 0x141 /* "Unexpected connected state" */);
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        this.channel.setConnectionState(connected, clientId);
    }
    process(message, local, localOpMetadata) {
        const safeTelemetryProps = (0, telemetry_utils_1.extractSafePropertiesFromMessage)(message);
        // On op process, tombstone error is logged in garbage collector. So, set "checkTombstone" to false when calling
        // "verifyNotClosed" which logs tombstone errors. Throw error if tombstoned and throwing on load is configured.
        this.verifyNotClosed("process", false /* checkTombstone */, safeTelemetryProps);
        if (this.tombstoned && this.gcThrowOnTombstoneUsage) {
            throw new telemetry_utils_1.DataCorruptionError("Context is tombstoned! Call site [process]", safeTelemetryProps);
        }
        this.summarizerNode.recordChange(message);
        if (this.loaded) {
            return this.channel?.process(message, local, localOpMetadata);
        }
        else {
            (0, core_utils_1.assert)(!local, 0x142 /* "local store channel is not loaded" */);
            (0, core_utils_1.assert)(this.pending !== undefined, 0x23d /* "pending is undefined" */);
            this.pending.push(message);
            this.thresholdOpsCounter.sendIfMultiple("StorePendingOps", this.pending.length);
        }
    }
    processSignal(message, local) {
        this.verifyNotClosed("processSignal");
        // Signals are ignored if the store is not yet loaded
        if (!this.loaded) {
            return;
        }
        this.channel?.processSignal(message, local);
    }
    getQuorum() {
        return this.parentContext.getQuorum();
    }
    getAudience() {
        return this.parentContext.getAudience();
    }
    /**
     * Returns a summary at the current sequence number.
     * @param fullTree - true to bypass optimizations and force a full summary tree
     * @param trackState - This tells whether we should track state from this summary.
     * @param telemetryContext - summary data passed through the layers for telemetry purposes
     */
    async summarize(fullTree = false, trackState = true, telemetryContext) {
        return this.summarizerNode.summarize(fullTree, trackState, telemetryContext);
    }
    async summarizeInternal(fullTree, trackState, telemetryContext) {
        await this.realize();
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        const summarizeResult = await this.channel.summarize(fullTree, trackState, telemetryContext);
        // Wrap dds summaries in .channels subtree.
        (0, index_js_1.wrapSummaryInChannelsTree)(summarizeResult);
        const pathPartsForChildren = [runtime_definitions_1.channelsTreeName];
        // Add data store's attributes to the summary.
        const { pkg } = await this.getInitialSnapshotDetails();
        const isRoot = await this.isRoot();
        const attributes = createAttributes(pkg, isRoot);
        (0, runtime_utils_1.addBlobToSummary)(summarizeResult, index_js_1.dataStoreAttributesBlobName, JSON.stringify(attributes));
        // If we are not referenced, mark the summary tree as unreferenced. Also, update unreferenced blob
        // size in the summary stats with the blobs size of this data store.
        if (!this.summarizerNode.isReferenced()) {
            summarizeResult.summary.unreferenced = true;
            summarizeResult.stats.unreferencedBlobSize = summarizeResult.stats.totalBlobSize;
        }
        // Add loadingGroupId to the summary
        if (this.loadingGroupId !== undefined) {
            summarizeResult.summary.groupId = this.loadingGroupId;
        }
        return {
            ...summarizeResult,
            id: this.id,
            pathPartsForChildren,
        };
    }
    /**
     * Returns the data used for garbage collection. This includes a list of GC nodes that represent this data store
     * including any of its child channel contexts. Each node has a set of outbound routes to other GC nodes in the
     * document.
     * If there is no new data in this data store since the last summary, previous GC data is used.
     * If there is new data, the GC data is generated again (by calling getGCDataInternal).
     * @param fullGC - true to bypass optimizations and force full generation of GC data.
     */
    async getGCData(fullGC = false) {
        return this.summarizerNode.getGCData(fullGC);
    }
    /**
     * Generates data used for garbage collection. This is called when there is new data since last summary. It
     * realizes the data store and calls into each channel context to get its GC data.
     * @param fullGC - true to bypass optimizations and force full generation of GC data.
     */
    async getGCDataInternal(fullGC = false) {
        await this.realize();
        (0, core_utils_1.assert)(this.channel !== undefined, 0x143 /* "Channel should not be undefined when running GC" */);
        return this.channel.getGCData(fullGC);
    }
    /**
     * After GC has run, called to notify the data store of routes used in it. These are used for the following:
     *
     * 1. To identify if this data store is being referenced in the document or not.
     *
     * 2. To determine if it needs to re-summarize in case used routes changed since last summary.
     *
     * 3. These are added to the summary generated by the data store.
     *
     * 4. To notify child contexts of their used routes. This is done immediately if the data store is loaded.
     * Else, it is done when realizing the data store.
     *
     * 5. To update the timestamp when this data store or any children are marked as unreferenced.
     *
     * @param usedRoutes - The routes that are used in this data store.
     */
    updateUsedRoutes(usedRoutes) {
        // Update the used routes in this data store's summarizer node.
        this.summarizerNode.updateUsedRoutes(usedRoutes);
        /**
         * Store the used routes to update the channel if the data store is not loaded yet. If the used routes changed
         * since the previous run, the data store will be loaded during summarize since the used state changed. So, it's
         * safe to only store the last used routes.
         */
        this.lastUsedRoutes = usedRoutes;
        // If we are loaded, call the channel so it can update the used routes of the child contexts.
        // If we are not loaded, we will update this when we are realized.
        if (this.loaded) {
            this.updateChannelUsedRoutes();
        }
    }
    /**
     * @deprecated There is no replacement for this, its functionality is no longer needed at this layer.
     * It will be removed in a future release, sometime after 2.0.0-internal.8.0.0
     *
     * Similar capability is exposed with from/to string paths instead of handles via @see addedGCOutboundRoute
     *
     * Called when a new outbound reference is added to another node. This is used by garbage collection to identify
     * all references added in the system.
     * @param srcHandle - The handle of the node that added the reference.
     * @param outboundHandle - The handle of the outbound node that is referenced.
     */
    addedGCOutboundReference(srcHandle, outboundHandle) {
        // By default, skip this call since the ContainerRuntime will detect the outbound route directly.
        if (this.mc.config.getBoolean(index_js_2.detectOutboundRoutesViaDDSKey) === true) {
            // Note: The ContainerRuntime code will check this same setting to avoid double counting.
            this.parentContext.addedGCOutboundReference?.(srcHandle, outboundHandle);
        }
    }
    /**
     * (Same as @see addedGCOutboundReference, but with string paths instead of handles)
     *
     * Called when a new outbound reference is added to another node. This is used by garbage collection to identify
     * all references added in the system.
     *
     * @param fromPath - The absolute path of the node that added the reference.
     * @param toPath - The absolute path of the outbound node that is referenced.
     */
    addedGCOutboundRoute(fromPath, toPath) {
        this.parentContext.addedGCOutboundReference?.({ absolutePath: fromPath }, { absolutePath: toPath });
    }
    /**
     * Updates the used routes of the channel and its child contexts. The channel must be loaded before calling this.
     * It is called in these two scenarios:
     * 1. When the used routes of the data store is updated and the data store is loaded.
     * 2. When the data store is realized. This updates the channel's used routes as per last GC run.
     */
    updateChannelUsedRoutes() {
        (0, core_utils_1.assert)(this.loaded, 0x144 /* "Channel should be loaded when updating used routes" */);
        (0, core_utils_1.assert)(this.channel !== undefined, 0x145 /* "Channel should be present when data store is loaded" */);
        // If there is no lastUsedRoutes, GC has not run up until this point.
        if (this.lastUsedRoutes === undefined) {
            return;
        }
        // Remove the route to this data store, if it exists.
        const usedChannelRoutes = this.lastUsedRoutes.filter((id) => {
            return id !== "/" && id !== "";
        });
        this.channel.updateUsedRoutes(usedChannelRoutes);
    }
    /**
     * @deprecated 0.18.Should call request on the runtime directly
     */
    async request(request) {
        const runtime = await this.realize();
        return runtime.request(request);
    }
    submitMessage(type, content, localOpMetadata) {
        this.verifyNotClosed("submitMessage");
        (0, core_utils_1.assert)(!!this.channel, 0x146 /* "Channel must exist when submitting message" */);
        // Summarizer clients should not submit messages.
        this.identifyLocalChangeInSummarizer("DataStoreMessageSubmittedInSummarizer", type);
        this.parentContext.submitMessage(type, content, localOpMetadata);
    }
    /**
     * This is called from a SharedSummaryBlock that does not generate ops but only wants to be part of the summary.
     * It indicates that there is data in the object that needs to be summarized.
     * We will update the latestSequenceNumber of the summary tracker of this
     * store and of the object's channel.
     *
     * @param address - The address of the channel that is dirty.
     *
     */
    setChannelDirty(address) {
        this.verifyNotClosed("setChannelDirty");
        // Get the latest sequence number.
        const latestSequenceNumber = this.deltaManager.lastSequenceNumber;
        this.summarizerNode.invalidate(latestSequenceNumber);
        const channelSummarizerNode = this.summarizerNode.getChild(address);
        if (channelSummarizerNode) {
            channelSummarizerNode.invalidate(latestSequenceNumber); // TODO: lazy load problem?
        }
    }
    /**
     * Submits the signal to be sent to other clients.
     * @param type - Type of the signal.
     * @param content - Content of the signal.
     * @param targetClientId - When specified, the signal is only sent to the provided client id.
     */
    submitSignal(type, content, targetClientId) {
        this.verifyNotClosed("submitSignal");
        (0, core_utils_1.assert)(!!this.channel, 0x147 /* "Channel must exist on submitting signal" */);
        return this.parentContext.submitSignal(type, content, targetClientId);
    }
    /**
     * This is called by the data store channel when it becomes locally visible indicating that it is ready to become
     * globally visible now.
     */
    makeLocallyVisible() {
        (0, core_utils_1.assert)(this.channel !== undefined, 0x2cf /* "undefined channel on datastore context" */);
        this.makeLocallyVisibleFn();
    }
    async bindRuntime(channel, existing) {
        if (this.channel) {
            throw new Error("Runtime already bound");
        }
        (0, core_utils_1.assert)(!this.detachedRuntimeCreation, 0x148 /* "Detached runtime creation on runtime bind" */);
        (0, core_utils_1.assert)(this.pkg !== undefined, 0x14a /* "Undefined package path" */);
        if (existing) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            const pending = this.pending;
            // Apply all pending ops
            for (const op of pending) {
                // Only process ops whose seq number is greater than snapshot sequence number from which it loaded.
                const seqNumber = this.baseSnapshotSequenceNumber ?? -1;
                if (op.sequenceNumber > seqNumber) {
                    channel.process(op, false, undefined /* localOpMetadata */);
                }
            }
            this.thresholdOpsCounter.send("ProcessPendingOps", pending.length);
        }
        else {
            (0, core_utils_1.assert)(this.pending?.length === 0, 0x8f4 /* no pending ops */);
            // Execute data store's entry point to make sure that for a local (aka detached from container) data store, the
            // entryPoint initialization function is called before the data store gets attached and potentially connected to
            // the delta stream, so it gets a chance to do things while the data store is still "purely local".
            // This preserves the behavior from before we introduced entryPoints, where the instantiateDataStore method
            // of data store factories tends to construct the data object (at least kick off an async method that returns
            // it); that code moved to the entryPoint initialization function, so we want to ensure it still executes
            // before the data store is attached.
            await channel.entryPoint.get();
        }
        this.pending = undefined;
        // And now mark the runtime active
        this.loaded = true;
        this.channel = channel;
        // Freeze the package path to ensure that someone doesn't modify it when it is
        // returned in packagePath().
        Object.freeze(this.pkg);
        /**
         * Update the used routes of the channel. If GC has run before this data store was realized, we will have
         * the used routes saved. So, this will ensure that all the child contexts have up-to-date used routes as
         * per the last time GC was run.
         * Also, this data store may have been realized during summarize. In that case, the child contexts need to
         * have their used routes updated to determine if its needs to summarize again and to add it to the summary.
         */
        this.updateChannelUsedRoutes();
    }
    async getAbsoluteUrl(relativeUrl) {
        if (this.attachState !== container_definitions_1.AttachState.Attached) {
            return undefined;
        }
        return this.parentContext.getAbsoluteUrl(relativeUrl);
    }
    /**
     * @deprecated Sets the datastore as root, for aliasing purposes: #7948
     * This method should not be used outside of the aliasing context.
     * It will be removed, as the source of truth for this flag will be the aliasing blob.
     */
    setInMemoryRoot() {
        this._isInMemoryRoot = true;
    }
    /**
     * @deprecated The functionality to get base GC details has been moved to summarizer node.
     */
    async getBaseGCDetails() {
        return {};
    }
    reSubmit(type, contents, localOpMetadata) {
        (0, core_utils_1.assert)(!!this.channel, 0x14b /* "Channel must exist when resubmitting ops" */);
        this.channel.reSubmit(type, contents, localOpMetadata);
    }
    rollback(type, contents, localOpMetadata) {
        if (!this.channel) {
            throw new Error("Channel must exist when rolling back ops");
        }
        if (!this.channel.rollback) {
            throw new Error("Channel doesn't support rollback");
        }
        this.channel.rollback(type, contents, localOpMetadata);
    }
    async applyStashedOp(contents) {
        if (!this.channel) {
            await this.realize();
        }
        (0, core_utils_1.assert)(!!this.channel, 0x14c /* "Channel must exist when rebasing ops" */);
        return this.channel.applyStashedOp(contents);
    }
    verifyNotClosed(callSite, checkTombstone = true, safeTelemetryProps = {}) {
        if (this.deleted) {
            const messageString = `Context is deleted! Call site [${callSite}]`;
            const error = new telemetry_utils_1.DataCorruptionError(messageString, safeTelemetryProps);
            this.mc.logger.sendErrorEvent({
                eventName: "GC_Deleted_DataStore_Changed",
                callSite,
            }, error);
            throw error;
        }
        if (this._disposed) {
            throw new Error(`Context is closed! Call site [${callSite}]`);
        }
        if (checkTombstone && this.tombstoned) {
            const messageString = `Context is tombstoned! Call site [${callSite}]`;
            const error = new telemetry_utils_1.DataCorruptionError(messageString, safeTelemetryProps);
            (0, index_js_2.sendGCUnexpectedUsageEvent)(this.mc, {
                eventName: "GC_Tombstone_DataStore_Changed",
                category: this.gcThrowOnTombstoneUsage ? "error" : "generic",
                gcTombstoneEnforcementAllowed: this.gcTombstoneEnforcementAllowed,
                callSite,
            }, this.pkg, error);
            if (this.gcThrowOnTombstoneUsage) {
                throw error;
            }
        }
    }
    /**
     * Summarizer client should not have local changes. These changes can become part of the summary and can break
     * eventual consistency. For example, the next summary (say at ref seq# 100) may contain these changes whereas
     * other clients that are up-to-date till seq# 100 may not have them yet.
     */
    identifyLocalChangeInSummarizer(eventName, type) {
        if (this.clientDetails.type !== index_js_1.summarizerClientType ||
            this.localChangesTelemetryCount <= 0) {
            return;
        }
        // Log a telemetry if there are local changes in the summarizer. This will give us data on how often
        // this is happening and which data stores do this. The eventual goal is to disallow local changes
        // in the summarizer and the data will help us plan this.
        this.mc.logger.sendTelemetryEvent({
            eventName,
            type,
            isSummaryInProgress: this.summarizerNode.isSummaryInProgress?.(),
            stack: (0, telemetry_utils_1.generateStack)(),
        });
        this.localChangesTelemetryCount--;
    }
    getCreateChildSummarizerNodeFn(id, createParam) {
        return (summarizeInternal, getGCDataFn) => this.summarizerNode.createChild(summarizeInternal, id, createParam, undefined /* config */, getGCDataFn);
    }
    deleteChildSummarizerNode(id) {
        this.summarizerNode.deleteChild(id);
    }
    async uploadBlob(blob, signal) {
        return this.parentContext.uploadBlob(blob, signal);
    }
}
exports.FluidDataStoreContext = FluidDataStoreContext;
FluidDataStoreContext.pendingOpsCountThreshold = 1000;
class RemoteFluidDataStoreContext extends FluidDataStoreContext {
    constructor(props) {
        super(props, true /* existing */, false /* isLocalDataStore */, () => {
            throw new Error("Already attached");
        });
        this.initialSnapshotDetailsP = new core_utils_1.LazyPromise(async () => {
            // Sequence number of the snapshot.
            let sequenceNumber;
            if (this.snapshotFetchRequired) {
                (0, core_utils_1.assert)(this.loadingGroupId !== undefined, 0x8f5 /* groupId should be present to fetch snapshot */);
                const snapshot = await this.runtime.getSnapshotForLoadingGroupId([this.loadingGroupId], [this.id]);
                this._baseSnapshot = snapshot.snapshotTree;
                sequenceNumber = snapshot.sequenceNumber;
                this.snapshotFetchRequired = false;
            }
            let tree = this.baseSnapshot;
            let isRootDataStore = true;
            if (!!tree && tree.blobs[index_js_1.dataStoreAttributesBlobName] !== undefined) {
                // Need to get through snapshot and use that to populate extraBlobs
                const attributes = await (0, driver_utils_1.readAndParse)(this.storage, tree.blobs[index_js_1.dataStoreAttributesBlobName]);
                let pkgFromSnapshot;
                // Use the snapshotFormatVersion to determine how the pkg is encoded in the snapshot.
                // For snapshotFormatVersion = "0.1" (1) or above, pkg is jsonified, otherwise it is just a string.
                const formatVersion = (0, index_js_1.getAttributesFormatVersion)(attributes);
                if (formatVersion < 1) {
                    pkgFromSnapshot =
                        attributes.pkg.startsWith('["') && attributes.pkg.endsWith('"]')
                            ? JSON.parse(attributes.pkg)
                            : [attributes.pkg];
                }
                else {
                    pkgFromSnapshot = JSON.parse(attributes.pkg);
                }
                this.pkg = pkgFromSnapshot;
                /**
                 * If there is no isRootDataStore in the attributes blob, set it to true. This will ensure that
                 * data stores in older documents are not garbage collected incorrectly. This may lead to additional
                 * roots in the document but they won't break.
                 */
                isRootDataStore = attributes.isRootDataStore ?? true;
                if ((0, index_js_1.hasIsolatedChannels)(attributes)) {
                    tree = tree.trees[runtime_definitions_1.channelsTreeName];
                    (0, core_utils_1.assert)(tree !== undefined, 0x1fe /* "isolated channels subtree should exist in remote datastore snapshot" */);
                }
            }
            (0, core_utils_1.assert)(this.pkg !== undefined, 0x8f6 /* The datastore context package should be defined */);
            return {
                pkg: this.pkg,
                isRootDataStore,
                snapshot: tree,
                sequenceNumber,
            };
        });
        this._baseSnapshot = props.snapshotTree;
        this.snapshotFetchRequired = !!props.snapshotTree?.omitted;
        this.runtime = props.parentContext.containerRuntime;
        if (props.snapshotTree !== undefined) {
            this.summarizerNode.updateBaseSummaryState(props.snapshotTree);
        }
    }
    async getInitialSnapshotDetails() {
        return this.initialSnapshotDetailsP;
    }
    /**
     * @see FluidDataStoreContext.getAttachData
     */
    getAttachData(includeGCData) {
        throw new Error("Cannot attach remote store");
    }
}
exports.RemoteFluidDataStoreContext = RemoteFluidDataStoreContext;
/**
 * Base class for detached & attached context classes
 */
class LocalFluidDataStoreContextBase extends FluidDataStoreContext {
    constructor(props) {
        super(props, props.snapshotTree !== undefined ? true : false /* existing */, true /* isLocalDataStore */, props.makeLocallyVisibleFn);
        this.initialSnapshotDetailsP = new core_utils_1.LazyPromise(async () => {
            let snapshot = this.snapshotTree;
            let attributes;
            let isRootDataStore = false;
            if (snapshot !== undefined) {
                // Get the dataStore attributes.
                // Note: storage can be undefined in special case while detached.
                attributes = await (0, index_js_1.getFluidDataStoreAttributes)(this.storage, snapshot);
                if ((0, index_js_1.hasIsolatedChannels)(attributes)) {
                    snapshot = snapshot.trees[runtime_definitions_1.channelsTreeName];
                    (0, core_utils_1.assert)(snapshot !== undefined, 0x1ff /* "isolated channels subtree should exist in local datastore snapshot" */);
                }
                if (this.pkg === undefined) {
                    this.pkg = JSON.parse(attributes.pkg);
                    // If there is no isRootDataStore in the attributes blob, set it to true. This ensures that data
                    // stores in older documents are not garbage collected incorrectly. This may lead to additional
                    // roots in the document but they won't break.
                    if (attributes.isRootDataStore ?? true) {
                        isRootDataStore = true;
                        this.setInMemoryRoot();
                    }
                }
            }
            (0, core_utils_1.assert)(this.pkg !== undefined, 0x152 /* "pkg should be available in local data store" */);
            return {
                pkg: this.pkg,
                isRootDataStore,
                snapshot,
            };
        });
        // Summarizer client should not create local data stores.
        this.identifyLocalChangeInSummarizer("DataStoreCreatedInSummarizer");
        this.snapshotTree = props.snapshotTree;
        if (props.isRootDataStore === true) {
            this.setInMemoryRoot();
        }
        this.createProps = props.createProps;
        this.attachListeners();
    }
    attachListeners() {
        if (this.attachState !== container_definitions_1.AttachState.Attached) {
            this.once("attaching", () => {
                (0, core_utils_1.assert)(this.attachState === container_definitions_1.AttachState.Detached, 0x14d /* "Should move from detached to attaching" */);
                this._attachState = container_definitions_1.AttachState.Attaching;
            });
            this.once("attached", () => {
                (0, core_utils_1.assert)(this.attachState === container_definitions_1.AttachState.Attaching, 0x14e /* "Should move from attaching to attached" */);
                this._attachState = container_definitions_1.AttachState.Attached;
            });
        }
    }
    /**
     * @see FluidDataStoreContext.getAttachData
     */
    getAttachData(includeGCData, telemetryContext) {
        (0, core_utils_1.assert)(this.channel !== undefined, 0x14f /* "There should be a channel when generating attach message" */);
        (0, core_utils_1.assert)(this.pkg !== undefined, 0x150 /* "pkg should be available in local data store context" */);
        const attachSummary = this.channel.getAttachSummary(telemetryContext);
        // Wrap dds summaries in .channels subtree.
        (0, index_js_1.wrapSummaryInChannelsTree)(attachSummary);
        // Add data store's attributes to the summary.
        const attributes = createAttributes(this.pkg, this.isInMemoryRoot());
        (0, runtime_utils_1.addBlobToSummary)(attachSummary, index_js_1.dataStoreAttributesBlobName, JSON.stringify(attributes));
        if (includeGCData) {
            const gcData = this.channel.getAttachGCData?.(telemetryContext);
            if (gcData !== undefined) {
                (0, runtime_utils_1.addBlobToSummary)(attachSummary, runtime_definitions_1.gcDataBlobKey, JSON.stringify(gcData));
            }
        }
        // Add loadingGroupId to the summary
        if (this.loadingGroupId !== undefined) {
            attachSummary.summary.groupId = this.loadingGroupId;
        }
        return {
            attachSummary,
            type: this.pkg[this.pkg.length - 1],
        };
    }
    async getInitialSnapshotDetails() {
        return this.initialSnapshotDetailsP;
    }
    /**
     * A context should only be marked as deleted when its a remote context.
     * Session Expiry at the runtime level should have closed the container creating the local data store context
     * before delete is even possible. Session Expiry is at 30 days, and sweep is done 36+ days later from the time
     * it was unreferenced. Thus the sweeping container should have loaded from a snapshot and thus creating a remote
     * context.
     */
    delete() {
        // TODO: GC:Validation - potentially prevent this from happening or asserting. Maybe throw here.
        (0, index_js_2.sendGCUnexpectedUsageEvent)(this.mc, {
            eventName: "GC_Deleted_DataStore_Unexpected_Delete",
            message: "Unexpected deletion of a local data store context",
            category: "error",
            gcTombstoneEnforcementAllowed: undefined,
        }, this.pkg);
        super.delete();
    }
}
exports.LocalFluidDataStoreContextBase = LocalFluidDataStoreContextBase;
/**
 * context implementation for "attached" data store runtime.
 * Various workflows (snapshot creation, requests) result in .realize() being called
 * on context, resulting in instantiation and attachment of runtime.
 * Runtime is created using data store factory that is associated with this context.
 */
class LocalFluidDataStoreContext extends LocalFluidDataStoreContextBase {
    constructor(props) {
        super(props);
    }
}
exports.LocalFluidDataStoreContext = LocalFluidDataStoreContext;
/**
 * Detached context. Data Store runtime will be attached to it by attachRuntime() call
 * Before attachment happens, this context is not associated with particular type of runtime
 * or factory, i.e. it's package path is undefined.
 * Attachment process provides all missing parts - package path, data store runtime, and data store factory
 */
class LocalDetachedFluidDataStoreContext extends LocalFluidDataStoreContextBase {
    constructor(props) {
        super(props);
        this.detachedRuntimeCreation = true;
        this.channelToDataStoreFn = props.channelToDataStoreFn;
    }
    async attachRuntime(registry, dataStoreChannel) {
        (0, core_utils_1.assert)(this.detachedRuntimeCreation, 0x154 /* "runtime creation is already attached" */);
        this.detachedRuntimeCreation = false;
        (0, core_utils_1.assert)(this.channelP === undefined, 0x155 /* "channel deferral is already set" */);
        this.channelP = Promise.resolve()
            .then(async () => {
            const factory = registry.IFluidDataStoreFactory;
            const factory2 = await this.factoryFromPackagePath();
            (0, core_utils_1.assert)(factory2 === factory, 0x156 /* "Unexpected factory for package path" */);
            await super.bindRuntime(dataStoreChannel, false /* existing */);
            (0, core_utils_1.assert)(!(await this.isRoot()), 0x8f7 /* there are no more createRootDataStore() kind of APIs! */);
            return dataStoreChannel;
        })
            .catch((error) => {
            this.mc.logger.sendErrorEvent({ eventName: "AttachRuntimeError" }, error);
            // The following two lines result in same exception thrown.
            // But we need to ensure that this.channelDeferred.promise is "observed", as otherwise
            // out UT reports unhandled exception
            throw error;
        });
        return this.channelToDataStoreFn(await this.channelP);
    }
    async getInitialSnapshotDetails() {
        if (this.detachedRuntimeCreation) {
            throw new Error("Detached Fluid Data Store context can't be realized! Please attach runtime first!");
        }
        return super.getInitialSnapshotDetails();
    }
}
exports.LocalDetachedFluidDataStoreContext = LocalDetachedFluidDataStoreContext;
//# sourceMappingURL=dataStoreContext.js.map