"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.compatBehaviorAllowsGCMessageType = exports.dataStoreNodePathOnly = exports.urlToGCNodePath = exports.unpackChildNodesGCDetails = exports.getGCDataFromSnapshot = exports.concatGarbageCollectionData = exports.cloneGCData = exports.concatGarbageCollectionStates = exports.generateSortedGCState = exports.shouldAllowGcSweep = exports.getGCVersion = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const runtime_definitions_1 = require("@fluidframework/runtime-definitions");
function getGCVersion(metadata) {
    if (!metadata) {
        // Force to 0/disallowed in prior versions
        return 0;
    }
    return metadata.gcFeature ?? 0;
}
exports.getGCVersion = getGCVersion;
/**
 * Indicates whether Sweep is allowed for this document based on the persisted GC Feature Matrix and current gcGeneration.
 * This applies to the entire Sweep Phase the same - both Tombstone Enforcement (i.e. should loading a Tombstone fail?) and Deletion.
 *
 * In order to protect old documents that were created at a time when known bugs exist that violate GC's invariants
 * such that enforcing GC Sweep would cause legitimate data loss, the container author may increment the generation value for Sweep
 * such that containers created with a different value will not be subjected to GC Sweep.
 *
 * If no generation is provided, Sweep will be enabled for all documents.
 *
 * For backwards compatibility, the current generation value is also compared against the persisted gcTombstoneGeneration if present.
 *
 * @param featureMatrix - The GC Feature Matrix, containing the persisted generation value
 * @param currentGeneration - The current app-provided gcGeneration value
 * @returns true if GC Sweep should be allowed for this document
 */
function shouldAllowGcSweep(featureMatrix, currentGeneration) {
    // If no Generation value is provided for this session, default to true
    if (currentGeneration === undefined) {
        return true;
    }
    // tombstoneGeneration is the predecessor and needs to be supported for back-compat reasons
    const targetGeneration = featureMatrix.tombstoneGeneration ?? featureMatrix.gcGeneration;
    return currentGeneration === targetGeneration;
}
exports.shouldAllowGcSweep = shouldAllowGcSweep;
/**
 * Sorts the given GC state as per the id of the GC nodes. It also sorts the outbound routes array of each node.
 */
function generateSortedGCState(gcState) {
    const sortableArray = Object.entries(gcState.gcNodes);
    sortableArray.sort(([a], [b]) => a.localeCompare(b));
    const sortedGCState = { gcNodes: {} };
    for (const [nodeId, nodeData] of sortableArray) {
        nodeData.outboundRoutes.sort();
        sortedGCState.gcNodes[nodeId] = nodeData;
    }
    return sortedGCState;
}
exports.generateSortedGCState = generateSortedGCState;
/**
 * Concatenates the given GC states and returns the concatenated GC state.
 */
function concatGarbageCollectionStates(gcState1, gcState2) {
    const combinedGCNodes = {};
    for (const [nodeId, nodeData] of Object.entries(gcState1.gcNodes)) {
        combinedGCNodes[nodeId] = {
            outboundRoutes: Array.from(nodeData.outboundRoutes),
            unreferencedTimestampMs: nodeData.unreferencedTimestampMs,
        };
    }
    for (const [nodeId, nodeData] of Object.entries(gcState2.gcNodes)) {
        let combineNodeData = combinedGCNodes[nodeId];
        if (combineNodeData === undefined) {
            combineNodeData = {
                outboundRoutes: Array.from(nodeData.outboundRoutes),
                unreferencedTimestampMs: nodeData.unreferencedTimestampMs,
            };
        }
        else {
            // Validate that same node doesn't have different unreferenced timestamp.
            if (nodeData.unreferencedTimestampMs !== undefined &&
                combineNodeData.unreferencedTimestampMs !== undefined) {
                (0, core_utils_1.assert)(nodeData.unreferencedTimestampMs === combineNodeData.unreferencedTimestampMs, 0x5d7 /* Two entries for the same GC node with different unreferenced timestamp */);
            }
            combineNodeData = {
                outboundRoutes: [
                    ...new Set([...nodeData.outboundRoutes, ...combineNodeData.outboundRoutes]),
                ],
                unreferencedTimestampMs: nodeData.unreferencedTimestampMs ?? combineNodeData.unreferencedTimestampMs,
            };
        }
        combinedGCNodes[nodeId] = combineNodeData;
    }
    return { gcNodes: combinedGCNodes };
}
exports.concatGarbageCollectionStates = concatGarbageCollectionStates;
/**
 * Helper function that clones the GC data.
 * @param gcData - The GC data to clone.
 * @returns a clone of the given GC data.
 */
function cloneGCData(gcData) {
    const clonedGCNodes = {};
    for (const [id, outboundRoutes] of Object.entries(gcData.gcNodes)) {
        clonedGCNodes[id] = Array.from(outboundRoutes);
    }
    return {
        gcNodes: clonedGCNodes,
    };
}
exports.cloneGCData = cloneGCData;
/**
 * Concatenates the given GC data and returns the concatenated GC data.
 */
function concatGarbageCollectionData(gcData1, gcData2) {
    const combinedGCData = cloneGCData(gcData1);
    for (const [id, routes] of Object.entries(gcData2.gcNodes)) {
        if (combinedGCData.gcNodes[id] === undefined) {
            combinedGCData.gcNodes[id] = Array.from(routes);
        }
        else {
            const combinedRoutes = [...routes, ...combinedGCData.gcNodes[id]];
            combinedGCData.gcNodes[id] = [...new Set(combinedRoutes)];
        }
    }
    return combinedGCData;
}
exports.concatGarbageCollectionData = concatGarbageCollectionData;
/**
 * Gets the base garbage collection state from the given snapshot tree. It contains GC state, deleted nodes and
 * tombstones. The GC state may be written into multiple blobs. Merge the GC state from all such blobs into one.
 */
async function getGCDataFromSnapshot(gcSnapshotTree, readAndParseBlob) {
    let rootGCState = { gcNodes: {} };
    let tombstones;
    let deletedNodes;
    for (const key of Object.keys(gcSnapshotTree.blobs)) {
        // Update deleted nodes blob.
        if (key === runtime_definitions_1.gcDeletedBlobKey) {
            deletedNodes = await readAndParseBlob(gcSnapshotTree.blobs[key]);
            continue;
        }
        // Update tombstone blob.
        if (key === runtime_definitions_1.gcTombstoneBlobKey) {
            tombstones = await readAndParseBlob(gcSnapshotTree.blobs[key]);
            continue;
        }
        // Skip blobs that do not start with the GC prefix.
        if (!key.startsWith(runtime_definitions_1.gcBlobPrefix)) {
            continue;
        }
        const blobId = gcSnapshotTree.blobs[key];
        if (blobId === undefined) {
            continue;
        }
        const gcState = await readAndParseBlob(blobId);
        (0, core_utils_1.assert)(gcState !== undefined, 0x5d8 /* GC blob missing from snapshot */);
        // Merge the GC state of this blob into the root GC state.
        rootGCState = concatGarbageCollectionStates(rootGCState, gcState);
    }
    return { gcState: rootGCState, tombstones, deletedNodes };
}
exports.getGCDataFromSnapshot = getGCDataFromSnapshot;
/**
 * Helper function that unpacks the GC details of the children from a given node's GC details.
 * @param gcDetails - The GC details of a node.
 * @returns A map of GC details of each children of the the given node.
 */
function unpackChildNodesGCDetails(gcDetails) {
    const childGCDetailsMap = new Map();
    // If GC data is not available, bail out.
    if (gcDetails.gcData === undefined) {
        return childGCDetailsMap;
    }
    const gcNodes = gcDetails.gcData.gcNodes;
    for (const [id, outboundRoutes] of Object.entries(gcNodes)) {
        // Skip self-node since only children GC data is to be generated.
        if (id === "/") {
            continue;
        }
        (0, core_utils_1.assert)(id.startsWith("/"), 0x5d9 /* node id should always be an absolute route */);
        const childId = id.split("/")[1];
        let childGCNodeId = id.slice(childId.length + 1);
        // GC node id always begins with "/". Handle the special case where a child's id in the parent's GC nodes is
        // of format `/root`. In this case, the childId is root and childGCNodeId is "". Make childGCNodeId = "/".
        if (childGCNodeId === "") {
            childGCNodeId = "/";
        }
        let childGCDetails = childGCDetailsMap.get(childId);
        if (childGCDetails === undefined) {
            childGCDetails = { gcData: { gcNodes: {} }, usedRoutes: [] };
        }
        // gcData should not undefined as its always at least initialized as  empty above.
        (0, core_utils_1.assert)(childGCDetails.gcData !== undefined, 0x5da /* Child GC data should have been initialized */);
        childGCDetails.gcData.gcNodes[childGCNodeId] = [...new Set(outboundRoutes)];
        childGCDetailsMap.set(childId, childGCDetails);
    }
    if (gcDetails.usedRoutes === undefined) {
        return childGCDetailsMap;
    }
    // Remove the node's self used route, if any, and generate the children used routes.
    const usedRoutes = gcDetails.usedRoutes.filter((route) => route !== "" && route !== "/");
    for (const route of usedRoutes) {
        (0, core_utils_1.assert)(route.startsWith("/"), 0x5db /* Used route should always be an absolute route */);
        const childId = route.split("/")[1];
        const childUsedRoute = route.slice(childId.length + 1);
        const childGCDetails = childGCDetailsMap.get(childId);
        (0, core_utils_1.assert)(childGCDetails?.usedRoutes !== undefined, 0x5dc /* This should have be initialized when generate GC nodes above */);
        childGCDetails.usedRoutes.push(childUsedRoute);
        childGCDetailsMap.set(childId, childGCDetails);
    }
    return childGCDetailsMap;
}
exports.unpackChildNodesGCDetails = unpackChildNodesGCDetails;
/**
 * Trims the leading and trailing slashes from the given string.
 * @param str - A string that may contain leading and / or trailing slashes.
 * @returns A new string without leading and trailing slashes.
 */
function trimLeadingAndTrailingSlashes(str) {
    return str.replace(/^\/+|\/+$/g, "");
}
/** Reformats a request URL to match expected format for a GC node path */
function urlToGCNodePath(url) {
    return `/${trimLeadingAndTrailingSlashes(url.split("?")[0])}`;
}
exports.urlToGCNodePath = urlToGCNodePath;
/**
 * Pulls out the first path segment and formats it as a GC Node path
 * e.g. "/dataStoreId/ddsId" yields "/dataStoreId"
 */
function dataStoreNodePathOnly(subDataStorePath) {
    return `/${subDataStorePath.split("/")[1]}`;
}
exports.dataStoreNodePathOnly = dataStoreNodePathOnly;
/**
 * Utility to implement compat behaviors given an unknown message type
 * The parameters are typed to support compile-time enforcement of handling all known types/behaviors
 *
 * @param _unknownGCMessageType - Typed as never to ensure all known types have been
 * handled before calling this function (e.g. in a switch statement).
 * @param compatBehavior - Typed redundantly with CompatModeBehavior to ensure handling is added when updating that type
 */
function compatBehaviorAllowsGCMessageType(_unknownGCMessageType, compatBehavior) {
    // undefined defaults to same behavior as "FailToProcess"
    return compatBehavior === "Ignore";
}
exports.compatBehaviorAllowsGCMessageType = compatBehaviorAllowsGCMessageType;
//# sourceMappingURL=gcHelpers.js.map