/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { assert } from "@fluidframework/core-utils";
import { createChildLogger } from "@fluidframework/telemetry-utils";
function isGroupContents(opContents) {
    return opContents?.type === OpGroupingManager.groupedBatchOp;
}
export class OpGroupingManager {
    constructor(config, logger) {
        this.config = config;
        this.logger = createChildLogger({ logger, namespace: "OpGroupingManager" });
    }
    groupBatch(batch) {
        if (!this.shouldGroup(batch)) {
            return batch;
        }
        if (batch.content.length >= 1000) {
            this.logger.sendTelemetryEvent({
                eventName: "GroupLargeBatch",
                length: batch.content.length,
                threshold: this.config.opCountThreshold,
                reentrant: batch.hasReentrantOps,
                referenceSequenceNumber: batch.content[0].referenceSequenceNumber,
            });
        }
        for (const message of batch.content) {
            if (message.metadata) {
                const keys = Object.keys(message.metadata);
                assert(keys.length < 2, 0x5dd /* cannot group ops with metadata */);
                assert(keys.length === 0 || keys[0] === "batch", 0x5de /* unexpected op metadata */);
            }
        }
        const serializedContent = JSON.stringify({
            type: OpGroupingManager.groupedBatchOp,
            contents: batch.content.map((message) => ({
                contents: message.contents === undefined ? undefined : JSON.parse(message.contents),
                metadata: message.metadata,
                compression: message.compression,
            })),
        });
        const groupedBatch = {
            ...batch,
            content: [
                {
                    localOpMetadata: undefined,
                    metadata: undefined,
                    referenceSequenceNumber: batch.content[0].referenceSequenceNumber,
                    contents: serializedContent,
                    type: OpGroupingManager.groupedBatchOp,
                },
            ],
        };
        return groupedBatch;
    }
    ungroupOp(op) {
        if (!isGroupContents(op.contents)) {
            return [op];
        }
        const messages = op.contents.contents;
        let fakeCsn = 1;
        return messages.map((subMessage) => ({
            ...op,
            clientSequenceNumber: fakeCsn++,
            contents: subMessage.contents,
            metadata: subMessage.metadata,
            compression: subMessage.compression,
        }));
    }
    shouldGroup(batch) {
        return (
        // Grouped batching must be enabled
        this.config.groupedBatchingEnabled &&
            // The number of ops in the batch must surpass the configured threshold
            batch.content.length >= this.config.opCountThreshold &&
            // Support for reentrant batches must be explicitly enabled
            (this.config.reentrantBatchGroupingEnabled || batch.hasReentrantOps !== true));
    }
}
OpGroupingManager.groupedBatchOp = "groupedBatch";
//# sourceMappingURL=opGroupingManager.js.map