/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ITelemetryLoggerExt, LoggingError } from "@fluidframework/telemetry-utils";
import { ITelemetryBaseProperties } from "@fluidframework/core-interfaces";
import { Deferred, IPromiseTimer, IPromiseTimerResult } from "@fluidframework/core-utils";
import { IAckSummaryResult, INackSummaryResult, IBroadcastSummaryResult, ISummarizeResults, ISummarizeHeuristicData, ISubmitSummaryOptions, SubmitSummaryResult, SummarizeResultPart, ISummaryCancellationToken, SubmitSummaryFailureData, IRefreshSummaryAckOptions } from "./summarizerTypes.js";
import { IClientSummaryWatcher } from "./summaryCollection.js";
export type raceTimerResult<T> = {
    result: "done";
    value: T;
} | {
    result: IPromiseTimerResult["timerResult"];
} | {
    result: "cancelled";
};
/** Helper function to wait for a promise or PromiseTimer to elapse. */
export declare function raceTimer<T>(promise: Promise<T>, timer: Promise<IPromiseTimerResult>, cancellationToken?: ISummaryCancellationToken): Promise<raceTimerResult<T>>;
export type SummarizeReason = 
/**
 * Attempt to summarize after idle timeout has elapsed.
 * Idle timer restarts whenever an op is received. So this
 * triggers only after some amount of time has passed with
 * no ops being received.
 */
"idle"
/**
 * Attempt to summarize after a maximum time since last
 * successful summary has passed. This measures time since
 * last summary ack op was processed.
 */
 | "maxTime"
/**
 * Attempt to summarize after a maximum number of ops have
 * passed since the last successful summary. This compares
 * op sequence numbers with the reference sequence number
 * of the summarize op corresponding to the last summary
 * ack op.
 */
 | "maxOps"
/**
 * Special case to attempt to summarize one last time before the
 * summarizer client closes itself. This is to prevent cases where
 * the summarizer client never gets a chance to summarize, because
 * there are too many outstanding ops and/or parent client cannot
 * stay connected long enough for summarizer client to catch up.
 */
 | "lastSummary"
/** On-demand summary requested with specified reason. */
 | `onDemand;${string}`
/** Enqueue summarize attempt with specified reason. */
 | `enqueue;${string}`;
declare const summarizeErrors: {
    /**
     * Error encountered while generating the summary tree, uploading
     * it to storage, or submitting the op. It could be a result of
     * the client becoming disconnected while generating or an actual error.
     */
    readonly submitSummaryFailure: "Error while generating, uploading, or submitting summary";
    /**
     * The summaryAckWaitTimeout time has elapsed before receiving the summarize op
     * sent by this summarize attempt. It is expected to be broadcast quickly.
     */
    readonly summaryOpWaitTimeout: "Timeout while waiting for summarize op broadcast";
    /**
     * The summaryAckWaitTimeout time has elapsed before receiving either a
     * summaryAck or summaryNack op from the server in response to this
     * summarize attempt. It is expected that the server should respond.
     */
    readonly summaryAckWaitTimeout: "Timeout while waiting for summaryAck/summaryNack op";
    /**
     * The server responded with a summaryNack op, thus rejecting this
     * summarize attempt.
     */
    readonly summaryNack: "Server rejected summary via summaryNack op";
    readonly disconnect: "Summary cancelled due to summarizer or main client disconnect";
};
export declare const getFailMessage: (errorCode: keyof typeof summarizeErrors) => string;
export declare class SummarizeResultBuilder {
    readonly summarySubmitted: Deferred<SummarizeResultPart<SubmitSummaryResult, SubmitSummaryFailureData>>;
    readonly summaryOpBroadcasted: Deferred<SummarizeResultPart<IBroadcastSummaryResult>>;
    readonly receivedSummaryAckOrNack: Deferred<SummarizeResultPart<IAckSummaryResult, INackSummaryResult>>;
    /**
     * Fails one or more of the three results as per the passed params.
     * If submit fails, all three results fail.
     * If op broadcast fails, only op broadcast result and ack nack result fails.
     * If ack nack fails, only ack nack result fails.
     */
    fail(message: string, error: any, submitFailureResult?: SubmitSummaryFailureData, nackSummaryResult?: INackSummaryResult): void;
    build(): ISummarizeResults;
}
/**
 * Errors type for errors hit during summary that may be retriable.
 */
export declare class RetriableSummaryError extends LoggingError {
    readonly retryAfterSeconds?: number | undefined;
    readonly canRetry: boolean;
    constructor(message: string, retryAfterSeconds?: number | undefined, props?: ITelemetryBaseProperties);
}
/**
 * This class generates and tracks a summary attempt.
 */
export declare class SummaryGenerator {
    private readonly pendingAckTimer;
    private readonly heuristicData;
    private readonly submitSummaryCallback;
    private readonly successfulSummaryCallback;
    private readonly refreshLatestSummaryCallback;
    private readonly summaryWatcher;
    private readonly logger;
    private readonly summarizeTimer;
    constructor(pendingAckTimer: IPromiseTimer, heuristicData: ISummarizeHeuristicData, submitSummaryCallback: (options: ISubmitSummaryOptions) => Promise<SubmitSummaryResult>, successfulSummaryCallback: () => void, refreshLatestSummaryCallback: (options: IRefreshSummaryAckOptions) => Promise<void>, summaryWatcher: Pick<IClientSummaryWatcher, "watchSummary">, logger: ITelemetryLoggerExt);
    /**
     * Generates summary and listens for broadcast and ack/nack.
     * Returns true for ack, false for nack, and undefined for failure or timeout.
     * @param reason - reason for summarizing
     * @param options - refreshLatestAck to fetch summary ack info from server,
     * fullTree to generate tree without any summary handles even if unchanged
     */
    summarize(summaryOptions: ISubmitSummaryOptions, resultsBuilder?: SummarizeResultBuilder): ISummarizeResults;
    private summarizeCore;
    private addSummaryDataToTelemetryProps;
    private summarizeTimerHandler;
    dispose(): void;
}
export {};
//# sourceMappingURL=summaryGenerator.d.ts.map