/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { MockContainerRuntimeFactoryForReconnection, MockContainerRuntimeForReconnection, } from "@fluidframework/test-runtime-utils";
import { TypedEventEmitter } from "@fluid-internal/client-utils";
import { createChildLogger, raiseConnectedEvent, } from "@fluidframework/telemetry-utils";
import { MessageType, SummaryType, } from "@fluidframework/protocol-definitions";
import { mergeStats } from "@fluidframework/runtime-utils";
import { Summarizer, SummaryManager, SummaryCollection, RunWhileConnectedCoordinator, } from "../../summary/index.js";
export class MockContainerRuntimeFactoryForSummarizer extends MockContainerRuntimeFactoryForReconnection {
    createContainerRuntime(dataStoreRuntime, overrides) {
        const containerRuntime = new MockContainerRuntimeForSummarizer(dataStoreRuntime, this, this.runtimeOptions, overrides);
        this.runtimes.add(containerRuntime);
        return containerRuntime;
    }
}
const DefaultSummaryConfiguration = {
    state: "disableHeuristics",
    maxAckWaitTime: 3 * 60 * 1000,
    maxOpsSinceLastSummary: 7000,
    initialSummarizerDelayMs: 5 * 1000, // 5 secs.
};
export class MockContainerRuntimeForSummarizer extends MockContainerRuntimeForReconnection {
    constructor(dataStoreRuntime, factory, runtimeOptions = {}, overrides) {
        super(dataStoreRuntime, factory, runtimeOptions, overrides);
        this.logger = createChildLogger();
        this.disposed = false;
        this.deltaManager.on("op", (message) => {
            this.emit("op", message);
        });
        this.summarizerClientElection = new MockSummarizerClientElection(this.clientId);
        this.connectedState = new MockConnectedState(this.logger, this.clientId);
        this.summaryCollection = new SummaryCollection(this.deltaManager, this.logger);
        const summaryConfiguration = {
            ...DefaultSummaryConfiguration,
            ...runtimeOptions.summaryConfiguration,
        };
        this.summarizer = new Summarizer(this /* summarizerRuntime */, () => summaryConfiguration /* configurationGetter */, this /* ISummarizerInternalsProvider */, {} /* handleContext */, this.summaryCollection, async (runtime) => RunWhileConnectedCoordinator.create(runtime, () => this.deltaManager.active));
        this.summaryManager = new SummaryManager(this.summarizerClientElection, this.connectedState, this.summaryCollection, this.logger, async () => this.summarizer, new MockThrottler());
        this.summaryManager.start();
    }
    /** Prepare a SummaryNack to be sent by the server */
    prepareSummaryNack() {
        const contents = {
            summaryProposal: {
                summarySequenceNumber: this.deltaManager.lastSequenceNumber,
            },
        };
        this.deltaManager.prepareInboundResponse(MessageType.SummaryNack, contents);
    }
    /** Call on the Summarizer object to summarize */
    async summarize() {
        const result = this.summarizer.summarizeOnDemand({
            reason: "fuzzTest",
            retryOnFailure: false,
        });
        return Promise.all([
            result.summarySubmitted,
            result.summaryOpBroadcasted,
            result.receivedSummaryAckOrNack,
        ]);
    }
    async submitSummary(options) {
        const summaryMessage = {
            handle: "",
            head: "",
            message: "",
            parents: [],
        };
        const referenceSequenceNumber = this.deltaManager.lastSequenceNumber;
        const summarizeMessage = {
            type: MessageType.Summarize,
            clientSequenceNumber: 0,
            referenceSequenceNumber,
            contents: summaryMessage,
        };
        this.deltaManager.inbound.push({
            ...summarizeMessage,
            clientId: this.clientId,
            sequenceNumber: 0,
            minimumSequenceNumber: 0,
            timestamp: 0,
        });
        this.deltaManager.outbound.push([summarizeMessage]);
        const summaryStats = {
            ...mergeStats(),
            dataStoreCount: 1,
            summarizedDataStoreCount: 1,
            summaryNumber: 1,
        };
        return {
            stage: "submit",
            handle: "",
            clientSequenceNumber: -1,
            referenceSequenceNumber,
            minimumSequenceNumber: -1,
            submitOpDuration: 0,
            uploadDuration: 0,
            generateDuration: 0,
            forcedFullTree: false,
            summaryTree: {
                type: SummaryType.Tree,
                tree: {},
            },
            summaryStats,
        };
    }
    async refreshLatestSummaryAck(options) {
        // Do nothing
    }
    setConnectedState(value) {
        super.setConnectedState(value);
        this.connectedState.setConnectedState(value, this.clientId);
        this.summarizerClientElection.setClientId(this.clientId);
    }
    closeFn() {
        this.disposeFn();
    }
    disposeFn() {
        this.connected = false;
        this.disposed = true;
        this.summaryManager.dispose();
        this.summarizer.dispose();
        this.deltaManager.dispose();
    }
}
class MockSummarizerClientElection extends TypedEventEmitter {
    constructor(clientId) {
        super();
        this.setClientId(clientId);
    }
    setClientId(clientId) {
        this.electedClientId = clientId;
        this.electedParentId = clientId;
    }
}
class MockConnectedState extends TypedEventEmitter {
    constructor(logger, clientId) {
        super();
        this.logger = logger;
        this.clientId = clientId;
        this.connected = false;
    }
    setConnectedState(connected, clientId) {
        if (this.connected === connected) {
            return;
        }
        this.clientId = clientId;
        raiseConnectedEvent(this.logger, this, connected, clientId);
    }
}
class MockThrottler {
    constructor() {
        this.getDelay = () => 0;
        this.numAttempts = 3;
        this.delayWindowMs = 0;
        this.maxDelayMs = 0;
        this.delayFn = () => 0;
    }
}
//# sourceMappingURL=summarizerFuzzMocks.js.map